import pioneer.common.constants as Constants
from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.sources.filesource import FileSource, try_all_patterns
from pioneer.das.api.loaders import pickle_loader

import numpy as np
import pandas as pd
import os
import re
import yaml
import glob


class DirSource(FileSource):
    """Loads a list of files from a directory."""

    def __init__(self, path, pattern=None, sort=True, loader=None, check_timestamps=True):
        super(DirSource, self).__init__(path, pattern, sort, loader)
        self.path = path
        self.nb_data_per_pkl_file = 1
        self.member_cached = None

        self.files, self.time_of_issues, self.timestamps, self.nb_data_per_pkl_file = self.get_files()

        if check_timestamps:
            self._check_timestamps_consistency()

    def get_files(self):

        files = []

        timestamps = None
        time_of_issues = None
        nb_data_per_pkl_file = 1

        for fullpath in glob.glob(f'{self.path}/*'):
            name = fullpath.split('/')[-1]
            if self.pattern is None:
                match, self.pattern = try_all_patterns(name)
            else:
                match = re.match(self.pattern, name)
            if match:
                groups = match.groups()
                if self.sort and not groups:
                    raise ValueError('no groups')
                if self.sort and groups:
                    sample = (int(groups[0]), fullpath)
                else:
                    sample = fullpath
                files.append(sample)

            # read config for multiple rows per .pkl file (high fps sensors)
            elif(name == Constants.CONFIG_YML_PATTERN):
                with open(fullpath) as f:
                    data_yaml = yaml.safe_load(f)
                    nb_data_per_pkl_file = data_yaml['nb_vectors_in_pkl']
                
            # read timestamps
            elif(name == Constants.TIMESTAMPS_CSV_PATTERN):
                with open(fullpath) as f:
                    sensor_ts = pd.read_csv(f, delimiter=" ", dtype='u8', header=None).values

                    timestamps = sensor_ts[:,0]
                    # check if ts is always go up, for imu data (more than 1 data per pkl file) the test is not complet -> to improve
                    if len(timestamps)>2 and (np.min(np.diff(timestamps.astype(np.int64)))<0):
                        LoggingManager.instance().warning('Timestamps are not strictly increasing for datasource file {}'.format(self.path))

                    if sensor_ts.shape[1] > 1:
                        time_of_issues = sensor_ts[:,1]
                    else:
                        time_of_issues = timestamps
        
        if self.sort:
            files.sort()
            files = [s[1] for s in files]
            
        return files, time_of_issues, timestamps, nb_data_per_pkl_file

    def __del__(self):
        pass

    def __len__(self):
        return self.timestamps.shape[0]

    def __getitem__(self, index):
        if index >= len(self):
            raise IndexError(f'For datasource {os.path.basename(self.path)} index {index} >= len {len(self)}')
        if self.nb_data_per_pkl_file == 1:
            return self.loader(self.files[index])
        else:
            if(index<0):
                index = len(self) + index
                if(index < 0):
                    raise IndexError('index {} < 0'.format(index))
            member = self.files[index//self.nb_data_per_pkl_file]
            # no reload
            if self.member_cached is None or self.member_cached!=member:
                self.data_cached = self.loader(member)
                self.member_cached=member

            return self.data_cached[index%self.nb_data_per_pkl_file] # compatible with array and struct array

    def _get_nb_timestamps_and_files_or_rows(self):
        if(self.nb_data_per_pkl_file == 1):
            nfiles = len(self.files)
        else:
            if len(self.files) == 0:
                nfiles = 0
            else:
                nfiles = (len(self.files)-1)*self.nb_data_per_pkl_file + self.loader(self.files[-1]).shape[0]
        return len(self), nfiles

    def _check_timestamps_consistency(self):
        nts, nfiles = self._get_nb_timestamps_and_files_or_rows()

        if nfiles != nts:
            n = min(nts, nfiles)
            LoggingManager.instance().warning('The number of timestamps and data files are '
                            'different for sensor %s (nfiles: %d != nts: %d). '
                            'Keeping the %d first timestamps and files'
                            %(self.path, nfiles, nts, n))
            self.timestamps = self.timestamps[:n]

            if(self.nb_data_per_pkl_file == 1):
                self.files = self.files[:n]
            else:
                if n%self.nb_data_per_pkl_file == 0:
                    self.files = self.files[:int(n/self.nb_data_per_pkl_file)]
                else:
                    # on va conserver un fichier à la fin qui ne sera pas utilisé en entier
                    self.files = self.files[:n//self.nb_data_per_pkl_file+1]

            nts, nfiles = self._get_nb_timestamps_and_files_or_rows()
            assert nfiles == nts

    def get(self, name, loader=None):
        if loader is None:
            loader = pickle_loader
        try:
            data = loader(name)
        except KeyError:
            data = None
        return data


import pioneer.common.constants as Constants
from pioneer.das.api.datasources import DataSource
from pioneer.das.api.loaders import pickle_loader, txt_loader, image_loader

import re
from six import string_types

def _endswith(s, patterns):
    assert isinstance(s, string_types)
    if isinstance(patterns, string_types):
        patterns = [patterns,]
    return any([s.endswith(p) for p in patterns])

def try_all_patterns(name):
    for p in [Constants.NUMBERED_PICKLE_PATTERN,
              Constants.NUMBERED_PNG_PATTERN,
              Constants.NUMBERED_JPG_PATTERN]:
        match = re.match(p, name)
        if match:
            return match, p
    return None, None

class FileSource(DataSource):
    """Base class for loading files from a dataset. Will be used to
        load a list of echoes from pickle files contained in a folder or a tar
        archive. Subclases must be used.
    """

    def __init__(self, path, pattern = None, sort=True, loader=None):
        self.path = path
        self.pattern = pattern
        self.sort = sort
        self._loader = loader
        self.files = []

    @property
    def loader(self):
        if self._loader is None:
            if _endswith(self.pattern, '.pkl'):
                self._loader = pickle_loader
            elif _endswith(self.pattern, ['.png', '.jpg']):
                self._loader = image_loader
            elif _endswith(self.pattern, '.txt'):
                self._loader = txt_loader
            else:
                raise RuntimeError('could not guess loader from pattern extension')
        return self._loader

    def get_timestamps(self):
        raise NotImplementedError()



from pioneer.common import platform, linalg, IoU3d
from pioneer.common.logging_manager import LoggingManager
from pioneer.common import platform as platform_utils
from pioneer.das.api import categories
from pioneer.das.api.samples.sample import Sample

from transforms3d import euler
from typing import Iterable, Optional

import copy
import numpy as np
import warnings
warnings.simplefilter('once', DeprecationWarning)


class Box3d(Sample):

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super(Box3d, self).__init__(index, datasource, virtual_raw, virtual_ts)
        self._label_source = categories.get_source(platform_utils.parse_datasource_name(self.datasource.label)[2])

    def get_centers(self) -> np.ndarray:
        return self.raw['data']['c']

    def get_dimensions(self) -> np.ndarray:
        return self.raw['data']['d']

    def get_rotations(self) -> np.ndarray:
        return self.raw['data']['r']

    def get_confidences(self) -> Iterable[Optional[float]]:
        return self.raw.get('confidence', [None for _ in range(len(self))])

    def get_category_numbers(self) -> Iterable[int]:
        return self.raw['data']['classes']

    def get_categories(self) -> Iterable[str]:
        categorie_labels = []
        for category_number in self.get_category_numbers():
            labels = categories.CATEGORIES[self._label_source]
            label = labels.get(str(category_number), {'name':'?'})['name']
            categorie_labels.append(label)
        return categorie_labels

    def get_ids(self) -> Iterable[Optional[int]]:
        return self.raw['data']['id']

    def __len__(self) -> int:
        return self.get_centers().shape[0]

    def set_transform(self, transform:np.ndarray) -> 'Box3d':
        if transform is None: return self

        centers = self.get_centers()
        rotations = self.get_rotations()

        raw = copy.deepcopy(self.raw)

        for i in range(len(self)):
            box_transform = np.eye(4)
            box_transform[:3, :3] = euler.euler2mat(*rotations[i])
            box_transform[:3, 3] = centers[i,:]
            box_to_transform = transform @ box_transform
            raw['data']['c'][i,:] = box_to_transform[:3, 3]
            raw['data']['r'][i,:] = euler.mat2euler(box_to_transform)

        return Box3d(self.index, self.datasource, raw, self.timestamp)

    def set_referential(self, referential:str, ignore_orientation:bool=False, reference_ts:int=-1, dtype=np.float64) -> 'Box3d':
        referential = platform.referential_name(referential)
        if self.datasource.sensor.name == referential: return self
        transform = self.compute_transform(referential, ignore_orientation, reference_ts, dtype)
        return self.set_transform(transform)
    

    ### Legacy section ###

    def attributes(self):
        warnings.warn("Box3d.attributes() is deprecated.", DeprecationWarning)
        if 'attributes' in self.raw:
            return self.raw['attributes']
        LoggingManager.instance().warning(f"There are no 'attributes' for that sample {self.datasource.label}.")
        return None
    
    def dynamics(self):
        warnings.warn("Box3d.dynamics() is deprecated.", DeprecationWarning)
        if 'dynamics' in self.raw:
            return self.raw['dynamics']
        LoggingManager.instance().warning(f"There are no 'dynamics' for that sample {self.datasource.label}.")
        return None

    def confidences(self):
        warnings.warn("Box3d.confidences() is deprecated, use Box3d.get_confidences() instead.", DeprecationWarning)
        if 'confidence' not in self.raw: return
        return self.get_confidences()

    def label_names(self):
        warnings.warn("Box3d.label_names() is deprecated, use Box3d.get_categories() instead.", DeprecationWarning)
        return self.get_categories()

    def _mapto(self, tf=None):
        """Maps the box3d to the new referential, given a 4x4 matrix transformation"""
        warnings.warn("Box3d._mapto() is deprecated, use Box3d.set_transform() instead.", DeprecationWarning)
        return self.set_transform(tf).raw['data']
    
    def mapto(self, referential_or_ds:str=None, ignore_orientation:bool=False, reference_ts:int=-1, dtype=np.float64):
        """ Will map each box in another referential. """
        warnings.warn("Box3d.mapto() is deprecated, use Box3d.set_referential() instead.", DeprecationWarning)
        return self.set_referential(referential_or_ds, ignore_orientation, reference_ts, dtype).raw['data']

    def num_pts_in(self, pt_cloud, margin=0):
        """ Returns, for each box, the mask of those points from pt_cloud that are inside the box.
            Args:
                pt_cloud - (M,3)
                margin (optional) - positive float- increases the size of box

            Returns:
                mask - boolean (n_boxe,M)
        """
        warnings.warn("Box3d.num_pts_in() is deprecated.", DeprecationWarning)
        bbox = self.raw['data']
        nbpts = np.zeros((len(bbox), len(pt_cloud)), dtype=bool)
        for i in range(len(bbox)):
            tf_Localds_from_Box = np.eye(4)
            tf_Localds_from_Box[:3, :3] = euler.euler2mat(bbox['r'][i,0], bbox['r'][i,1], bbox['r'][i,2])
            tf_Localds_from_Box[:3, 3] = bbox['c'][i,:]
            aabb = np.vstack([-(bbox['d'][i,:]+margin) / 2.0,(bbox['d'][i,:]+margin) / 2.0])
            nbpts[i,:] = linalg.points_inside_box_mask(pt_cloud, aabb, linalg.tf_inv(tf_Localds_from_Box))
        return nbpts

    def set_angle_to_domain(self, domain=[0,2*np.pi]):
        """Will set the angles to a given domain"""
        warnings.warn("Box3d.set_angle_to_domain() is deprecated.", DeprecationWarning)

        bbox = np.copy(self.raw['data'])
        for i in range(len(bbox)):
            bbox['r'][i,:] = [linalg.map_angle_to_domain(bbox['r'][i,j], domain=domain) for j in range(3)]

        return bbox
    
    def compute_iou(self, box, return_max=False, map2yaw=None):
        """Compute the iou score between all the elements of self and of box.
            
            Return a matrix len(self), len(box) when row,col are indexed in the same order as self, box.

            If return_max=True: return only a single number for each element of self (the max value).

            Important note: By default the computation is performed in the sbg ref where only one angle (yaw) is not zero, unless
            map2yaw is provided (a callable) which brings all the boxes in a referential where only one rotation (yaw).

        """
        warnings.warn("Box3d.compute_iou() is deprecated.", DeprecationWarning)
        if map2yaw is not None:
            box0 = map2yaw(self)
            box1 = map2yaw(box)
        else:
            try: #must find either sbg or ENU.
                referential = platform.referential_name('sbgekinox_bcc')
                tf_TargetRef_from_Local = np.copy(self.datasource.sensor.map_to(referential))
                tf_TargetRef_from_Local[:3, :3] = tf_TargetRef_from_Local[:3, :3] @ self.orientation

            except:
                LoggingManager.instance().warning('IoU computation, falling back to ENU system transformation.')
                tf_TargetRef_from_Local = np.eye(4)
                tf_TargetRef_from_Local[:3, :3] = np.array([[0, 0, 1],
						                                    [-1, 0, 0],
						                                    [0, -1, 0]],
						                                    dtype=np.float).T
            box0 = self._mapto(tf=tf_TargetRef_from_Local)
            box1 = box._mapto(tf=tf_TargetRef_from_Local)
        
        Z0 = [box0['c'], box0['d'], 'z', box0['r'][:,2]]
        Z1 = [box1['c'], box1['d'], 'z', box1['r'][:,2]]
        matiou = IoU3d.matrixIoU(Z0=Z0, Z1=Z1)
        if return_max:
            return np.max(matiou, axis=1)
        else:
            return matiou


from pioneer.common.linalg import map_points, tf_inv
from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.datatypes import datasource_xyzit
from pioneer.das.api.samples.echo import Echo
from pioneer.das.api.samples.point_cloud import PointCloud

from typing import Any, Iterable

import copy
import numpy as np


try :
    import open3d
    OPEN3D_AVAILABLE = True
except:
    LoggingManager.instance().warning("Voxelisation will not work without installing open3d -> pip3 install open3d")
    OPEN3D_AVAILABLE = False

class VoxelMap(VirtualDatasource):
    """Merges multiple past points clouds in a single one, in the same present referential (mapping).
        Optionally, the merged point cloud can be voxelized (one averaged point per voxel, or volume unit).

        In order to work, the platform needs to have an egomotion_provider.
    """
    def __init__(self, reference_sensor:str, dependencies:list, memory:int=5, skip:int=1, voxel_size:float=0.0):
        """Constructor
            Args:
                reference_sensor (str): The name of the sensor (e.g. 'ouster64_bfc').
                dependencies (list): A list of the datasource names. 
                    The only element should be a point cloud datasource (e.g. 'ouster64_bfc_xyzit')
                memory (int): The number of past frames to merge.
                skip (int): If higher than 1, frames will be skipped. For example, with memory=10 and skip=2, 
                    the frames N=0,-2,-4,-6,-8,-10 will be merged. The present frame (N=0) is always included.
                voxel_size (float): If greater than 0, the merged point cloud will be voxelized.
        """
        self.has_rgb = 'xyzit-rgb' in dependencies[0]
        ds_type = 'xyzit-voxmap' if not self.has_rgb else 'xyzit-voxmap-rgb'
        super(VoxelMap, self).__init__(ds_type, dependencies, None)
        self.reference_sensor = reference_sensor
        self.original_point_cloud_datasource = dependencies[0]
        self.memory = memory
        self.skip = skip
        self.voxel_size = voxel_size
        self.local_cache = None
        

    def get_at_timestamp(self, timestamp):
        sample = self.datasources[self.original_point_cloud_datasource].get_at_timestamp(timestamp)
        return self[int(np.round(sample.index))]

    @staticmethod
    def stack_point_cloud(stack, point_cloud):
        return np.vstack([stack, point_cloud])

    def clear_cache(self):
        self.local_cache = None

    def voxelize(self, pc):

        pc_open3d = open3d.geometry.PointCloud()
        pc_open3d.points = open3d.utility.Vector3dVector(pc[:,[0,1,2]])
        pc_open3d.colors = open3d.utility.Vector3dVector(pc[:,[3,3,3]])

        try:
            down_sampled = pc_open3d.voxel_down_sample(voxel_size=self.voxel_size)
        except:
            OPEN3D_AVAILABLE = False
            return pc_vox

        xyz_vox = np.asarray(down_sampled.points)
        int_vox = np.asarray(down_sampled.colors)[:,0]

        pc_vox = np.empty((xyz_vox.shape[0], pc.shape[1]))
        pc_vox[:,[0,1,2]] = xyz_vox
        pc_vox[:,3] = int_vox
        if pc.shape[0] > 0:
            pc_vox[:,4] = pc[:,4].max()

        if self.has_rgb:
            pc_open3d.colors = open3d.utility.Vector3dVector(pc[:,[6,7,8]])
            down_sampled_rgb = pc_open3d.voxel_down_sample(voxel_size=self.voxel_size)
            pc_vox[:,[6,7,8]] = np.asarray(down_sampled_rgb.colors)[:,[0,1,2]]

        return pc_vox

    def __getitem__(self, key:Any):

        #TODO: if multiple point cloud datasources in dependencies, we could merge them.

        min_key = key - self.memory
        min_key = max([0, min_key])

        samples:Iterable[PointCloud] = self.datasources[self.original_point_cloud_datasource][min_key:key+1]

        nb_features = 1 if not self.has_rgb else 4
        pc_map = np.empty((0,5+nb_features))

        cached_indices = []
        if self.local_cache is not None:
            pc_map = self.local_cache

            keep = np.where(
                (pc_map[:,5] >= min_key) &\
                (pc_map[:,5] <= key) &\
                (pc_map[:,5] % self.skip == 0)
            )
            pc_map = pc_map[keep]
            cached_indices = np.unique(pc_map[:,5])

        
        for sample in samples:

            index = sample.index
            if index % self.skip and index != key:
                continue

            if index in cached_indices:
                continue #don't re-add what is already cached

            pc = np.empty((sample.size,5+nb_features))
            pc[:,[0,1,2]] = sample.get_point_cloud(referential='world', undistort=False)
            if isinstance(sample, Echo):
                pc[:,3] = sample.amplitudes
                pc[:,4] = sample.timestamps
            else:
                pc[:,3] = sample.get_field('i')
                pc[:,4] = sample.get_field('t')
            pc[:,5] = index

            if self.has_rgb:
                pc[:,6] = sample.get_field('r')
                pc[:,7] = sample.get_field('g')
                pc[:,8] = sample.get_field('b')

            pc_map = self.stack_point_cloud(pc_map, pc)

        self.local_cache = copy.deepcopy(pc_map)

        if self.voxel_size > 0 and OPEN3D_AVAILABLE:
            pc_map = self.voxelize(pc_map)

        to_world = samples[-1].compute_transform('world')
        to_sensor = tf_inv(to_world)
        pc_map[:,[0,1,2]] = map_points(to_sensor, pc_map[:,[0,1,2]])

        #package in das format
        dtype = datasource_xyzit() if not self.has_rgb else sample.raw.dtype
        raw = np.empty((pc_map.shape[0]), dtype=dtype)
        raw['x'] = pc_map[:,0]
        raw['y'] = pc_map[:,1]
        raw['z'] = pc_map[:,2]
        raw['i'] = pc_map[:,3]
        raw['t'] = pc_map[:,4]

        if self.has_rgb:
            raw['r'] = pc_map[:,6]
            raw['g'] = pc_map[:,7]
            raw['b'] = pc_map[:,8]

        return PointCloud(index=key, datasource=self, virtual_raw=raw, virtual_ts=samples[-1].timestamp)




from pioneer.das.api.interpolators import nearest_interpolator
from pioneer.das.api.samples.xyzv import XYZV
from pioneer.das.api.sensors.sensor3d import Sensor3D

from typing import Any, Dict, Tuple


class Radar(Sensor3D):
    def __init__(self, name: str, platform: 'Platform', factories:Dict[str, Tuple[Any, Any]] = {}):
        super().__init__(name, platform, 
            factories = {**factories,
                'xyzvi': (XYZV, nearest_interpolator), 
                'xyzvcfar': (XYZV, nearest_interpolator),
            })



from pioneer.common import platform
from pioneer.das.api.interpolators import from_float_index
from pioneer.das.api.datasources import AbstractDatasource
from pioneer.das.api.sources import FileSource

from collections import deque
from typing import Optional, List, Callable, Dict, Tuple, Any, Iterable, Union

import numbers
import numpy as np

class DatasourceWrapper(AbstractDatasource):

    def __init__(self, sensor:'Sensor', ds_type:str, ds:FileSource, sample_interp:Tuple[Any, Any], cache_size:int=100):
        super(DatasourceWrapper, self).__init__(sensor, ds_type)
        self.sensor = sensor
        self.ds = ds
        self.ds_type = ds_type
        self._cache = {}
        self._cache_keys = deque(maxlen = cache_size)
        self.sample_factory, self.interpolator = sample_interp
        self.use_cache = cache_size > 0
        
    def invalidate_caches(self):
        self._cache = {}
        self._cache_keys = deque(maxlen = self._cache_keys.maxlen)
        self._interpolated = {}
        self._interpolated_keys = deque(maxlen = self._cache_keys.maxlen)

    @property
    def interpolator(self):
        return self._interpolator

    @interpolator.setter
    def interpolator(self, interpolator):
        self._interpolator = interpolator
        self._interpolated = {}
        self._interpolated_keys = deque(maxlen = self._cache_keys.maxlen)

    @property
    def timestamps(self):
        return self.ds.timestamps

    @property
    def time_of_issues(self):
        return self.ds.time_of_issues

    def _add_interpolated_from_float_index(self, float_index:float):
        if self.interpolator is None:
            raise RuntimeError('Index is float {} : No inperpolator found'.format(float_index))
        raw = self.interpolator(self, float_index)
        self._interpolated_keys.append(float_index)
        v =  self._interpolated[float_index] = self.sample_factory(float_index, self.sensor[self.ds_type], raw, self.to_timestamp(float_index))
        return v

    def to_timestamp(self, float_index:float) -> int:
        i, t = from_float_index(float_index)
        if i == 0: i += 1
        ts_from = self.timestamps[i - 1]
        return int(ts_from + t * (self.timestamps[i] - ts_from))

    def _add_interpolated_from_timestamp(self, timestamp:int) -> float:
        float_index = self.to_float_index(timestamp)
        self._add_interpolated_from_float_index(float_index)
        return float_index

    def get_at_timestamp(self, timestamp:Union[int, Iterable], interpolator:Callable=None) -> Union['Sample', List['Sample']]:
        try: #is timestamp iterable?
            return [self.get_at_timestamp(t, interpolator) for t in timestamp]
        except:
            float_index = self.to_float_index(timestamp)
            if interpolator is None:
                return self[float_index]
            return self.sample_factory(float_index, self.sensor[self.ds_type], interpolator(self, float_index), timestamp)

    def __getitem__(self, key:Any) -> Union['Sample', List['Sample']]:

        def try_get_from_cache_or_make_room(i, cache, cache_keys):
            if not self.use_cache:
                return None
            n = len(cache_keys)
            if i not in cache_keys:
                if n >= cache_keys.maxlen:
                    k = cache_keys.popleft()
                    s = cache.pop(k)
                    del s
                return None
            if n > 1:
                cache_keys.append(cache_keys.popleft()) #make latest access the most recent
            return cache[i]

        def instantiate(i):

            i = self.to_positive_index(i)
               
            if i >= len(self):
                raise IndexError("Out of bound")
            
            v = try_get_from_cache_or_make_room(i, self._cache, self._cache_keys)

            if v is None:
                if self.use_cache:
                    self._cache_keys.append(i)
                    v = self._cache[i] = self.sample_factory(i, self.sensor[self.ds_type])
                else:
                    v = self.sample_factory(i, self.sensor[self.ds_type])
                if self.sensor.platform.is_live():
                    # When the platform is live the fifo could fill up before data is accessed, 
                    # so we must avoid relying on it to get Sample.raw, Sample.timestamp
                    v.virtual_raw = v.raw
                    v.virtual_ts = v.timestamp

            return v

            self._cache_keys.append(self._cache_keys.popleft()) #make latest access the most recent
            return self._cache[i]

        try: #is key iterable?
            return [self[index] for index in key]
        except:
            if isinstance(key, slice):
                return self[platform.slice_to_range(key, len(self))]
            elif isinstance(key, numbers.Integral):
                return instantiate(key)
            elif isinstance(key, numbers.Real) and float(key).is_integer():
                return instantiate(int(key))
            elif isinstance(key, numbers.Real) and not float(key).is_integer():
                v = try_get_from_cache_or_make_room(key, self._interpolated, self._interpolated_keys)
                if v is None:
                    v = self._add_interpolated_from_float_index(key)

                return v

        raise KeyError('Invalid key: {}'.format(key))

    @property
    def label(self):
        return "{}_{}".format(self.sensor.name, self.ds_type)

    def get_timestamp_slice(self, timestamp:int, half_window_size:Union[int,tuple]):
        """Get an index slice that represent a time interval.

        The time interval is parametrized by a timestamp and a half window
        size.

        Args:
            timestamp: The center of the time interval
            half_window_size: Half of the time interval. If tuple provided -> (start, end).

        Returns:
            slice -- The indices slice representing all the elements that fit
            in the time interval
        """
        try: #is 2-tuple?
            left = int(timestamp) + int(half_window_size[0])
            right = int(timestamp) + int(half_window_size[1])
        except:
            left = int(timestamp) - int(half_window_size)
            right = int(timestamp) + int(half_window_size)
        left_index = int(np.searchsorted(self.timestamps, left, side='left'))
        right_index = int(np.searchsorted(self.timestamps, right, side='right'))
        return slice(left_index, right_index, 1)




import pioneer.common.constants as Constants
from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.sources.filesource import FileSource, try_all_patterns
from pioneer.das.api.loaders import pickle_loader

from ruamel.std import zipfile

import multiprocessing
import numpy as np
import pandas as pd
import os
import re
import sys
import threading
import time
import traceback
import yaml

class ZipFileSource(FileSource):
    """Loads a list of files from a zip archive."""

    def __init__(self, path, pattern=None, sort=True, loader=None,
                 check_timestamps=True, mode="lock"):
        super(ZipFileSource, self).__init__(path, pattern, sort, loader)
        self.path = path
        self.member_cached = None
        
        self.lock = multiprocessing.RLock()
        self.tlock = threading.RLock()
        self.archive = None
        self.set_mode(mode)

        self.nb_data_per_pkl_file = 1

        self.files, self.time_of_issues, self.timestamps, self.nb_data_per_pkl_file = self.get_files(self.access)

        if check_timestamps:
            self._check_timestamps_consistency()

    def get_files(self, access):

        try:
            members = access( lambda archive: archive.infolist())
        except Exception as e:
            traceback.print_exc()
            raise e

        files = []

        timestamps = None
        time_of_issues = None
        nb_data_per_pkl_file = 1

        for m in members:
            name = m.filename
            if self.pattern is None:
                match, self.pattern = try_all_patterns(name)
            else:
                match = re.match(self.pattern, name)
            if match:
                groups = match.groups()
                if self.sort and not groups:
                    raise ValueError('no groups')
                if self.sort and groups:
                    sample = (int(groups[0]), m)
                else:
                    sample = m
                files.append(sample)

            # read config for multiple rows per .pkl file (high fps sensors)
            elif(name == Constants.CONFIG_YML_PATTERN):
                def f_CONFIG_YML_PATTERN(archive):
                    with archive.open(name) as stream:
                        data_yaml = yaml.safe_load(stream)
                        return data_yaml['nb_vectors_in_pkl']
                
                nb_data_per_pkl_file = access(f_CONFIG_YML_PATTERN)

            # read timestamps
            elif(name == Constants.TIMESTAMPS_CSV_PATTERN):
                def f_TIMESTAMPS_CSV_PATTERN(archive):
                    with archive.open(name) as stream:
                        # sensor_ts = np.loadtxt(stream, dtype='u8', delimiter=' ', ndmin=2)
                        sensor_ts = pd.read_csv(stream, delimiter=" ", dtype='u8', header=None).values

                        timestamps = sensor_ts[:,0]
                        # check if ts is always go up, for imu data (more than 1 data per pkl file) the test is not complet -> to improve
                        if len(timestamps)>2 and (np.min(np.diff(timestamps.astype(np.int64)))<0):
                            LoggingManager.instance().warning('Timestamps are not strictly increasing for datasource file {}'.format(self.path))

                        if sensor_ts.shape[1] > 1:
                            time_of_issues = sensor_ts[:,1]
                        else:
                            time_of_issues = timestamps
                        return time_of_issues, timestamps

                time_of_issues, timestamps = access(f_TIMESTAMPS_CSV_PATTERN)
         
        if self.sort:
            files.sort()
            files = [s[1] for s in files]
            
        return files, time_of_issues, timestamps, nb_data_per_pkl_file


    def set_mode(self, mode):

        def access_file(f):
            with zipfile.ZipFile(self.path, 'r') as archive:
                return f(archive)

        def access_raw(f):
            return f(self.archive)

        def access_lock(f):
            exception = None
            while True:
                try: # workaround for BadZipFile in multiprocessing mode
                    with self.lock:
                        with self.tlock:
                            return f(self.archive)
                except Exception as e:
                    if False:
                        sys.stdout.write(f"\rZipFileSource {os.path.basename(self.path)} read error, falling back to 'file' mode{' ' * 10}")
                        sys.stdout.flush()
                    try:
                        if self.archive is not None:
                            self.archive.close()
                            self.archive = None
                        self.files, self.time_of_issues, self.timestamps, self.nb_data_per_pkl_file = self.get_files(access_file)
                        rv =  access_file(f)
                        self.archive = zipfile.ZipFile(self.path, 'r')
                        return rv

                    except Exception as access_file_e:
                        self.archive = zipfile.ZipFile(self.path, 'r')
                        print(f'\rZipFileSource fallback failed: {access_file_e}, retrying infinitely')
                        time.sleep(0.1)

        if mode == "lock":
            self.archive = zipfile.ZipFile(self.path, 'r')
            self.access = access_lock
        elif mode == "file":
            self.archive = None
            self.access = access_file
        elif mode == "none":
            self.archive = zipfile.ZipFile(self.path, 'r')
            self.access = access_raw

    def __del__(self):
        if self.archive is not None:
            self.archive.close()

    def __len__(self):
        return self.timestamps.shape[0]

    def __getitem__(self, index):
        if index >= len(self):
            raise IndexError(f'For datasource {os.path.basename(self.path)} index {index} >= len {len(self)}')
        if self.nb_data_per_pkl_file == 1:
            member = self.files[index]
            data_bytes = self.access( lambda archive: archive.read(member))
            return self.loader(data_bytes)
        # pkl containing nb_data_per_pkl_file data rows
        else:
            if(index<0):
                index = len(self) + index
                if(index < 0):
                    raise IndexError('index {} < 0'.format(index))
            member = self.files[index//self.nb_data_per_pkl_file]
            # no reload
            if self.member_cached is None or self.member_cached!=member:
                data_bytes = self.access( lambda archive: archive.read(member))
                self.data_cached = self.loader(data_bytes)
                self.member_cached=member

            return self.data_cached[index%self.nb_data_per_pkl_file] # compatible with array and struct array

    def _get_nb_timestamps_and_files_or_rows(self):
        # nb timestamps
        nts = len(self)
        # nb files, or rows
        if(self.nb_data_per_pkl_file == 1):
            nfiles = len(self.files)
        else:
            # if empty no data
            if len(self.files) == 0:
                nfiles = 0
            # else, (N-1) x nb_data_per_pkl_file + nb rows in last file
            else:
                data_bytes = self.access(lambda archive: archive.read(self.files[-1]))
                nfiles = (len(self.files)-1)*self.nb_data_per_pkl_file + self.loader(data_bytes).shape[0]
        return nts, nfiles


    def _check_timestamps_consistency(self):
        nts, nfiles = self._get_nb_timestamps_and_files_or_rows()

        if nfiles != nts:
            n = min(nts, nfiles)
            LoggingManager.instance().warning('The number of timestamps and data files are '
                            'different for sensor %s (nfiles: %d != nts: %d). '
                            'Keeping the %d first timestamps and files'
                            %(self.path, nfiles, nts, n))
            self.timestamps = self.timestamps[:n]

            if(self.nb_data_per_pkl_file == 1):
                self.files = self.files[:n]
            else:
                if n%self.nb_data_per_pkl_file == 0:
                    self.files = self.files[:int(n/self.nb_data_per_pkl_file)]
                else:
                    # on va conserver un fichier à la fin qui ne sera pas utilisé en entier
                    self.files = self.files[:n//self.nb_data_per_pkl_file+1]

            nts, nfiles = self._get_nb_timestamps_and_files_or_rows()
            assert nfiles == nts

    def get(self, name, loader=None):
        if loader is None:
            loader = pickle_loader
        try:
            data_bytes = self.access(lambda archive: archive.read(archive.getinfo(name)))
            data = loader(data_bytes)
        except KeyError:
            data = None
        return data


import setuptools

def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

install_reqs = parse_requirements('requirements.txt')

setuptools.setup(
    name="pioneer_das_api", # Replace with your own username
    version="1.4.4",
    author="Leddartech",
    description="Pioneer's das api",
    packages=[
        'pioneer', 
        'pioneer.das', 
        'pioneer.das.api', 
        'pioneer.das.tests',
        'pioneer.das.api.samples', 
        'pioneer.das.api.samples.annotations',
        'pioneer.das.api.sensors', 
        'pioneer.das.api.sources',
        'pioneer.das.api.datasources',
        'pioneer.das.api.datasources.virtual_datasources',
        'pioneer.das.api.egomotion'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    dependency_links=["https://pioneer:yK6RUkhUCNHg3e1yxGT4@svleddar-gitlab.leddartech.local/api/v4/projects/481/packages/pypi/simple/pioneer-common"],
    install_requires=install_reqs,
    include_package_data=True
)


from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.datasources.virtual_datasources.bounding_box_metadata import BoundingBoxMetadata
from pioneer.das.api.datasources.virtual_datasources.cylindrical_projection import FlirCylindricalProjection
from pioneer.das.api.datasources.virtual_datasources.echoes_from_traces import Echoes_from_Traces
from pioneer.das.api.datasources.virtual_datasources.echoes_with_pulses import EchoesWithPulses
from pioneer.das.api.datasources.virtual_datasources.hdr_waveform import HDRWaveform
from pioneer.das.api.datasources.virtual_datasources.point_cloud_fusion import PointCloudFusion
from pioneer.das.api.datasources.virtual_datasources.rgb_cloud import RGBCloud
from pioneer.das.api.datasources.virtual_datasources.waveform_cloud import WaveformCloud
from pioneer.das.api.datasources.virtual_datasources.voxel_map import VoxelMap

VIRTUAL_DATASOURCE_FACTORY = {
    'bounding_box_metadata': BoundingBoxMetadata,
    'cylindrical_projection': FlirCylindricalProjection,
    'echoes_from_traces': Echoes_from_Traces,
    'echoes_with_pulses': EchoesWithPulses,
    'hdr_waveform': HDRWaveform,
    'point_cloud_fusion': PointCloudFusion, 
    'rgb_cloud': RGBCloud,
    'waveform_cloud': WaveformCloud,
    'voxel_map': VoxelMap,
}


from pioneer.das.api.samples.annotations import *
from pioneer.das.api.samples.echo import Echo
from pioneer.das.api.samples.fast_trace import FastTrace
from pioneer.das.api.samples.image import Image
from pioneer.das.api.samples.image_cylinder import ImageCylinder
from pioneer.das.api.samples.image_fisheye import ImageFisheye
from pioneer.das.api.samples.point_cloud import PointCloud
from pioneer.das.api.samples.sample import Sample
from pioneer.das.api.samples.trace import Trace
from pioneer.das.api.samples.xyzit import XYZIT
from pioneer.das.api.samples.xyzv import XYZV



#TODO: move this to pioneer.das.acquisition
from pioneer.das.api import platform

try:
    import folium #pip3 install folium
except:
    pass
import math
import matplotlib.pyplot as plt
import numpy as np
import os
import tqdm
import utm

def easting_northing_from_lat_long(latitude, longitude):
    easting, northing, _, _ = utm.from_latlon(latitude, longitude)
    return easting, northing

def distance_traj_step(easting, northing, t=None):
    d_e = np.diff(easting)
    d_n = np.diff(northing)
    if t is not None:
        d_t = np.diff(t)
    return (d_e**2 + d_n**2)**0.5/d_t

def get_trajectory(pfsynch:platform.Synchronized,
                    ref_ts_sensor:str='flir_bfc_img',
                    imu_nav:str='sbgekinox_bcc_navposvel'):
    
    '''simple: return easting, northing, points list, and time of the trajectory following the timestamps of ref_ts_sensor
    '''
    n = len(pfsynch)
    easting, northing, ts = [], [], []
    points = []
    for mu in tqdm.tqdm(range(n)):
        ref_ts = pfsynch[mu][ref_ts_sensor].timestamp
        imu = pfsynch.platform[imu_nav].get_at_timestamp(ref_ts).raw
        lati, longi = imu['latitude'], imu['longitude']
        eg, ng = easting_northing_from_lat_long(lati, longi)
        easting.append(eg)
        northing.append(ng)
        ts.append(ref_ts/1e6)
        points.append([lati, longi])
    
    return np.array(easting, dtype=np.float64), np.array(northing, dtype=np.float64), points, np.array(ts, dtype=np.float64)-ts[0]


def compute_neighbour_step_ratio(xt, yt, t, min_epsilon_precision=1e-5):
    step_ratio_norm = []
    step_ratio_norm.append(0)
    for i in range(1,len(xt)-1):
        d_t_l = np.abs(t[i-1]-t[i])
        d_t_r = np.abs(t[i+1]-t[i])

        d_xt_l = np.maximum(np.abs(xt[i-1]-xt[i])/d_t_l, min_epsilon_precision)
        d_xt_r = np.maximum(np.abs(xt[i+1]-xt[i])/d_t_r, min_epsilon_precision)
        
        d_yt_l = np.maximum(np.abs(yt[i-1]-yt[i])/d_t_l, min_epsilon_precision)
        d_yt_r = np.maximum(np.abs(yt[i+1]-yt[i])/d_t_r, min_epsilon_precision)

        step_ratio_xt = np.maximum(d_xt_l, d_xt_r) / np.minimum(d_xt_l, d_xt_r)
        step_ratio_yt = np.maximum(d_yt_l, d_yt_r) / np.minimum(d_yt_l, d_yt_r)
        
        step_ratio_norm.append((step_ratio_xt**2 + step_ratio_yt**2)**0.5)
    
    step_ratio_norm.append(0)
    return np.array(step_ratio_norm, dtype=np.float)


def compute_standard_score(x, seq_memory: int=200, start_at_zero: bool=True, outliers_threshold: float=100.0):

    '''return the standard score based on a memory sequence of certain length. 
    '''
    m = len(x)
    epsilon_ = 1e-4 # 0.1 mm precision
    z_score = []
    z_score.append(0)
    flag_outliers = np.zeros_like(x, dtype=bool)
    for mu in tqdm.tqdm(range(1, m)):
        a, b = np.maximum(mu - seq_memory, 0), mu
        
        if mu < seq_memory and not start_at_zero:
            z_score.append(0)
            continue
        
        window_seq = x[a:b][~flag_outliers[a:b]]
        
        # if mu > seq_memory and len(window_seq) < 0.25*seq_memory:
        #     z_score.append(0)
        #     continue
        
        seq_mean = np.mean(window_seq)
        seq_std = np.std(window_seq)

        z_ = np.abs((x[mu] - seq_mean)/(seq_std + epsilon_))

        if z_ > outliers_threshold:
            flag_outliers[mu] = 1

        z_score.append(np.copy(z_))
    
    return np.array(z_score)


def get_trajectory_standard_score(pfsynch:platform.Synchronized,
                                    ref_ts_sensor:str='flir_bfc_img',
                                    imu_nav:str='sbgekinox_bcc_navposvel',
                                    traj_seq_memory:int=200):
    
    '''estimation of the smoothness of a trajectory based on the standard score.
    '''

    easting, northing, _, t = get_trajectory(pfsynch, ref_ts_sensor, imu_nav)
    traj_step = distance_traj_step(easting, northing, t)
    z_scores = np.zeros_like(easting)
    z_scores[1:] = compute_standard_score(traj_step, traj_seq_memory-1, False)
    return z_scores


def get_trajectory_step_ratio(pfsynch:platform.Synchronized,
                                ref_ts_sensor:str='flir_bfc_img',
                                imu_nav:str='sbgekinox_bcc_navposvel',
                                traj_min_epsilon_precision:float=1e-6):

    '''estimation of the smoothness of the trajectory based on the ratio of left-right epsilons step
    '''
    easting, northing, _, t = get_trajectory(pfsynch, ref_ts_sensor, imu_nav)
    return compute_neighbour_step_ratio(easting, northing, t, traj_min_epsilon_precision)


def find_trajectory_jump(pfsynch:platform.Synchronized,
                            ref_ts_sensor:str='flir_bfc_img',
                            imu_nav:str='sbgekinox_bcc_navposvel',
                            traj_seq_memory:int=200,
                            traj_jump_threshold:float=15.5,
                            show_result:bool=True):
    
    '''Compute the list of ranges of intervall from pfsynch which are smooth according to traj_jump_threshold.
    '''
    
    print('Computing trajectory')
    easting, northing, points, t = get_trajectory(pfsynch, ref_ts_sensor, imu_nav)
    traj_step = distance_traj_step(easting, northing)

    print('Validate trajectory')
    z_scores = np.zeros_like(easting)
    z_scores[1:] = compute_standard_score(traj_step, traj_seq_memory-1, False)
    jump_flag = (z_scores > traj_jump_threshold).astype(bool)
    
    list_intervals = []
    ids = np.arange(len(jump_flag))[jump_flag]
    for mu in range(len(ids)):
        if mu == 0:
            list_intervals.append([0 , ids[mu]-1])
            continue
        if ids[mu]-ids[mu-1] >= traj_seq_memory:
            list_intervals.append([ids[mu-1], ids[mu]-1])
    
    if show_result:
        t = np.arange(len(easting))
        fig, ax = plt.subplots(2, 1, figsize=(9,10))
        fig.suptitle('Trajectory positions and jumps')
        ax[0].scatter(t, easting)
        ax[0].scatter(t[jump_flag], easting[jump_flag], label='jump flags')
        ax[0].legend()
        ax[0].set_xlabel('Frame number')
        ax[0].set_ylabel('Easting')

        ax[1].scatter(t, northing)
        ax[1].scatter(t[jump_flag], northing[jump_flag], label='jump flags')
        ax[1].legend()
        ax[1].set_xlabel('Frame number')
        ax[1].set_ylabel('Northing')
        plt.show()

        my_map = folium.Map(location=points[0],  zoom_start=15)
        folium.PolyLine(points).add_to(my_map)
        for mu in ids:
            folium.CircleMarker(
                    location=points[mu],
                    radius=5.5,
                    popup='IMU jump: '+ str(mu),
                    color='red',
                    fill=True,
                    fill_color='red'
                ).add_to(my_map)
        return jump_flag, list_intervals, my_map
    
    return jump_flag, list_intervals





if __name__ == '__main__':
    #example of use:

    #see this dataset:
    _dataset = '/nas/pixset/exportedDataset/20200610_195655_rec_dataset_quartier_pierre_exported'
    _ignore = ['radarTI_bfc']
    pf = platform.Platform(_dataset, ignore=_ignore)

    # get the platform synchronized:
    sync_labels = ['*ech*', '*_img*', '*_trr*', '*_trf*',' *_ftrr*', '*xyzit-*']
    interp_labels = ['*_xyzit', 'sbgekinox_*', 'peakcan_*', '*temp', '*_pos*', '*_agc*']

    synch = pf.synchronized(sync_labels=sync_labels, interp_labels=interp_labels, tolerance_us=1e3)

    flags, inters, my_map = find_trajectory_jump(synch,
                                            ref_ts_sensor='flir_bfc_img',
                                            imu_nav='sbgekinox_bcc_navposvel',
                                            traj_seq_memory=200,
                                            traj_jump_threshold=4.0,
                                            show_result=True)

    print('Intervals:', inters)

    








    

   
    





import numpy as np

def box2d():
    '''Fields are: 
        'x' (float): the top coordinate of the box in pixels
        'y' (float): the left coordinate of the box in pixels
        'h' (float): box height in pixels
        'w' (float): box width in pixels
        'r' (float): box rotation angle (anti-clockwise)
        'classes' (int): the object category number
        'id' (int): the object's instance unique id
        'flags' (int): miscellaneous infos
    '''
    return np.dtype([
        ('x', 'f4'),
        ('y', 'f4'),
        ('h', 'f4'),
        ('w', 'f4'),
        ('r','f4'),
        ('classes', 'u2'),
        ('id', 'u2'),
        ('flags', 'u2')
    ])

def box3d():
    '''Fields are: 
        'c' (float, float, float): center coordinates
        'd' (float, float, float): the box dimensions
        'r' (float, float, float): the Euler angles (rotations)
        'classes' (int): the object category number
        'id' (int): the object's instance unique id 
        'flags' (int): miscellaneous infos
    Coordinate system: +x is forward, +y is left and +z is up. 
    '''
    return np.dtype([
        ('c', 'f4', (3)),
        ('d', 'f4', (3)),
        ('r','f4', (3)),
        ('classes', 'u2'),
        ('id', 'u2'),
        ('flags', 'u2'),
    ])

def attributes():
    return np.dtype([
        ('occlusions', 'u2'),
        ('truncations', 'u2'),
        ('on the road','?'),
        ('vehicle activities', 'O'),
        ('human activities', 'O')
    ])

def box3d_dynamics():
    return np.dtype([
        ('vel', 'f4', (3)),
        ('acc', 'f4', (3)),
        ('angvel', 'f4'),
    ])

def seg3d():
    '''Fields are: 
        'classes' (int): the object category number
        'id' (int): the object's instance unique id 
    Note: A 3D segmentation array must have the same length and ordering as the corresponding point cloud array.
    '''
    return np.dtype([('classes', 'u2'), ('id', 'u2')])

def poly2d():
    '''Fields are: 
        'polygon': an (N,2) array with N vertices and 2 x,y coordinates (in pixel units)
        'classes' (int): the object category number
        'id' (int): the object's instance unique id
        'flags' (int): miscellaneous infos
    '''
    return np.dtype([
        ('polygon','O'),
        ('classes','u2'),
        ('id','u2'),
        ('flags','u2'),
    ])

def seg2d():
    '''Fields are: 
        'confidences': 2D array of values between 0 and 1
        'classes' (int): the object category number
    Note: the size of the 2D arrays should be the same as the sensor's resolution
    '''
    return np.dtype([
        ('confidences','O'),
        ('classes','u2'),
    ])

def datasource_xyzit():
    '''Fields are: 
        'x' (float): x coordinate (forward)
        'y' (float): y coordinate (left)
        'z' (float): z coordinate (up)
        'i' (int): intensity
        't' (int): timestamp
    '''
    return np.dtype([('x','f4'),('y','f4'),('z','f4'),('i','u2'),('t','u8')])
    
def datasource_xyzit_float_intensity():
    '''Fields are: 
        'x' (float): x coordinate (forward)
        'y' (float): y coordinate (left)
        'z' (float): z coordinate (up)
        'i' (float): intensity
        't' (int): timestamp
    '''
    return np.dtype([('x','f4'),('y','f4'),('z','f4'),('i','f4'),('t','u8')])

def lane():
    """Fields are: 
        'vertices': (N,3) numpy array of the 3D coordinates
        'type' (int): number to identify the type of lane (see das/lane_type.py)
    """
    return np.dtype([('vertices','O'),('type','u4')])


from pioneer.common import clouds, banks, misc
from pioneer.common.logging_manager import LoggingManager
from pioneer.common.types import calibration
from pioneer.das.api.interpolators import nearest_interpolator, linear_dict_of_float_interpolator, floor_interpolator
from pioneer.das.api.samples.point_cloud import PointCloud
from pioneer.das.api.sensors.sensor import Sensor
from pioneer.das.api.samples import Sample, Echo, Trace, FastTrace

import numpy as np
from typing import Union, Dict


class LCAx(Sensor):
    """ LeddarTech LCAx family of sensors, expects 'ech', 'sta' and 'cfg' datasources"""

    def __init__(self, name: str, platform: 'Platform'):
        super(LCAx, self).__init__(name
                                   , platform
                                   , {  'ech': (Echo, nearest_interpolator)
                                      , 'sta': (Sample, linear_dict_of_float_interpolator)
                                      , 'cfg': (Sample, floor_interpolator)
                                      , 'trr': (Trace, nearest_interpolator)
                                      , 'ftrr': (FastTrace, nearest_interpolator)
                                      , 'xyzit': (PointCloud, nearest_interpolator)
                                     }
                                   )

        self._cache = {}
        # intrinsics override
        self.specs = None
        self.temperature_slope = None
        self.temperature_reference = None
        self.time_base_delays = None
        self.static_noise = None
        self.angle_chart = None
        self.angle_chart_path = None
        self.angle_chart_ref_v_fov = None
        self.mirror_temp_compensation = None
        self.factor_correction = None
        self.projection = None
        self.config = {'reject_flags': [3, 3072]
        , 'extrema_amp' : 2**20 #2d images only
        , 'extrema_dist': 250 #2d images only
        , 'dist_reject_intervals': []
        , 'amp_reject_intervals': []}
        self.scan_direction = 'vertical'
        self.saturation_calibration = None
        self.orientation = np.array([[ 1,  0,  0],
                                     [ 0,  1,  0],
                                     [ 0,  0,  1]], 'f8')
        self.pcl_datasource = 'ech'


    def cache(self, specs: Dict[str, Union[int, float]]) -> Dict[str, np.ndarray]:
        """ caches computation that are shared for all 'ech' samples.
        
        Args:
            specs: a dict with keys, 'v'\: number of scan channels, 'h'\: the number of imaging channels,
                'v_fov'\: scan field of view, in degrees, 'h_fov'\: imaging field of view, in degrees

        Returns: 
            A dict with keys 'angles' (the projection angles), 'directions' (the projection directions) 
            and 'quad_directions' (the projection surface cloud directions)
        Note:
            If 'angle_chart' is defined in this sensor's intrinsics, the angle chart's angles is used to
            compute 'directions', but **not** 'quad_directions'.
            If 'mirror_temp_compensation' is defined in this sensor's intrisincs, 'angles_temp_lut' 
            and 'directions_temp_lut' will appear in cache and be used by apply_direction_corrections()

        """

        hashable = frozenset([(k,specs[k]) for k in ['v', 'h', 'v_fov', 'h_fov']]) #order invariant set

        if not hashable in self._cache:
            cache = {}

            if self.angle_chart is not None :
                specs = dict(specs)
                ## To be coherent with angle in chart table
                if self.angle_chart_ref_v_fov is not None : 
                    specs['v_fov'] = self.angle_chart_ref_v_fov

                if self.mirror_temp_compensation is not None:
                    angles_temp_lut = {}
                    directions_temp_lut = {}
                    quad_directions_temp_lut = {}
                    
                    temp_min, m, b = self.mirror_temp_compensation['t_min'], self.mirror_temp_compensation['slope'], self.mirror_temp_compensation['y0']

                    for t in np.arange(round(temp_min), 60, 0.5):
                        f = m*t + b
                        a = angles_temp_lut[t] = clouds.custom_v_angles(specs, factor = f, filename = self.angle_chart_path)
                        directions_temp_lut[t] = clouds.directions(a)
                        quad_directions_temp_lut[t] = clouds.custom_v_quad_directions(specs, factor = f, filename = self.angle_chart_path)

                    cache['angles_temp_lut'] = angles_temp_lut
                    cache['directions_temp_lut'] = directions_temp_lut
                    cache['quad_directions_temp_lut'] = quad_directions_temp_lut
                    
                cache['angles'] = clouds.custom_v_angles(specs, factor = self.factor_correction, filename = self.angle_chart_path)
                cache['directions'] = clouds.directions(cache['angles'])
                cache['quad_directions'] = clouds.custom_v_quad_directions(specs, factor = self.factor_correction, filename = self.angle_chart_path)

            else:
                if self.projection == 'direction_orthogonal':
                    cache['angles'] = clouds.angles(specs)
                    cache['directions'] = clouds.directions(cache['angles'],direction_f=getattr(clouds,self.projection))
                    cache['quad_directions'] = clouds.quad_directions(specs,direction_f=getattr(clouds,self.projection))
                elif self.projection == 'direction_carla':
                    cache['angles'] = clouds.angles(specs)
                    cache['directions'],cache['quad_directions'] = clouds.directions_orthogonal(specs)
                elif self.projection == 'direction_carla_pixell':
                    cache['angles'] = clouds.angles(specs)
                    cache['directions'],cache['quad_directions'] = clouds.directions_orthogonal_pixell(specs)
                elif self.projection == 'direction_spherical':
                    cache['angles'] = clouds.angles(specs)
                    cache['directions'] = clouds.directions(cache['angles'], direction_f=clouds.direction_spherical)
                    cache['quad_directions'] = clouds.quad_directions(specs, direction_f=clouds.direction_spherical)                    
                else: #default mode
                    cache['angles'] = clouds.angles(specs)
                    cache['directions'] = clouds.directions(cache['angles'], direction_f=clouds.direction)
                    cache['quad_directions'] = clouds.quad_directions(specs, direction_f=clouds.direction)
            
            cache['corrected_specs'] = specs
            
            self._cache[hashable] = cache
       
        return self._cache[hashable]

    def load_intrinsics(self, intrinsics_config: Union[str, dict]):
        """Load the LCAx intrinsics from a dictionnary like {'v': 8, 'h', 32, 'v_fov': 20, 'h_fov': 30}

        Args:
            intrinsics_config: The intrinsics configuration
        """
        if isinstance(intrinsics_config, dict):
            intrinsics_folder = intrinsics_config.get('folder', None)
            if intrinsics_folder is not None:
                super(LCAx, self).load_intrinsics(intrinsics_folder)

            specs_config = intrinsics_config.get('specs', None)
            if specs_config is not None:
                self.specs = banks.fill_specs(**intrinsics_config.get('specs', None))
            else:
                self.load_specs_from_cfg()
            self.angle_chart = intrinsics_config.get('angle_chart', None)
            if (self.angle_chart is not None) and (not isinstance(self.angle_chart,str)): ## Old platform still have path written in string directly
                self.angle_chart_path = self.angle_chart['path']
                self.angle_chart_ref_v_fov = self.angle_chart['ref_v_fov']
            else :
                self.angle_chart_path = self.angle_chart
            self.factor_correction = intrinsics_config.get('factor_correction', 1)
            self.mirror_temp_compensation = intrinsics_config.get('mirror_temp_compensation', None)
            self.projection = intrinsics_config.get('projection', None)
            self.adc_freq = float(intrinsics_config.get('adc_frequency', 100e6))
            temp_config = intrinsics_config.get('temperature', None)
            if temp_config:
                self.temperature_slope = temp_config['slope']
                self.temperature_reference = temp_config['reference']

            # self.load_static_noise_from_cfg()
            self.load_time_base_delays_from_cfg()

            saturation_calibration_config = intrinsics_config.get('saturation_calibration', None)
            if saturation_calibration_config is not None:
                self.saturation_calibration = calibration.SaturationCalibration(
                                                self.yml['intrinsics']['saturation_calibration']['distance_coefficient'],
                                                self.yml['intrinsics']['saturation_calibration']['amplitude_coefficients'])

        else:
            super(LCAx, self).load_intrinsics(intrinsics_config)

    def load_specs_from_cfg(self):
        try:
            cfg = self['cfg'][0].raw
            self.specs = banks.extract_specs(lambda n: cfg[n])
        except:
            LoggingManager.instance().warning("Unable to read the specs from the sensor {}".format(self.name))

    def load_time_base_delays_from_cfg(self):
        try:
            cfg = self['cfg'][0].raw
            self.time_base_delays = banks.extract_intrinsics_timebase_delays(lambda n: cfg[n])
        except:
            self.time_base_delays = 0 if not 'ftrr' in self else {'high':0, 'low':0}
            LoggingManager.instance().warning("Unable to read the time base delays from the sensor {}".format(self.name))
                
    def load_static_noise_from_cfg(self):
        try:
            cfg = self['cfg'][0].raw
            self.static_noise = banks.extract_intrinsics_static_noise(lambda n: cfg[n])
        except:
            self.static_noise = 0
            LoggingManager.instance().warning("Unable to read the static noise from the sensor {}".format(self.name))

    def get_corrected_projection_data(self, timestamp:Union[int, float], cache:dict, type:str='directions'):
        """ Returns temperature compensated projections directions or angles

        Args:
            timestamp:  the timestamp at which the sample is needed (use to obtain the right temperature)
            cache:      the cache to search in (obtained by calling LCAx.cache(...))
            type:       the type of directions (e.g) 'directions' or 'quad_directions'. 
                        **Must be present in the cache**

        Returns:
            the compensated directions if any, or the uncompensated direction otherwise
        """

        default_directions = cache[type]

        if self.mirror_temp_compensation is not None:
            try:
                temp = self.get_temperature_at(timestamp)
            except Exception as e:
                LoggingManager.instance().warning(f"Failed to apply temperature correction. Error: {str(e)}")
                return default_directions
            
            temp = min(max(temp, self.mirror_temp_compensation['t_min']), 60)
            temp_floor = np.floor(temp)
            if (temp - temp_floor) >= 0.5:
                temp = temp_floor + 0.5
            else:
                temp = temp_floor

            return cache[f'{type}_temp_lut'][temp]
            
        return default_directions
    
    def get_corrected_cloud(self, timestamp, cache, type, indices, distances, amplitudes=None, dtype=np.float64):
        """Returns the point-cloud (or quad-cloud) using the get_corrected_projection_data() method.

            Args:
                type: 'point_cloud' or 'quad_cloud'

        """
        if type == 'point_cloud':
            directions_ = self.get_corrected_projection_data(timestamp, cache, type='directions')
            return clouds.to_point_cloud(indices, distances, directions_, dtype)
        elif type == 'quad_cloud':
            quad_directions_ = self.get_corrected_projection_data(timestamp, cache, type='quad_directions')
            return clouds.to_quad_cloud(indices, distances, amplitudes, quad_directions_, cache['corrected_specs']['v'],cache['corrected_specs']['h'], dtype)
        else:
            raise KeyError('Can not compute the point cloud projection')
        
    def apply_distance_corrections(self, timestamp, indices, distances):
        """Applies calibration and temperature-related distances corrections"""
        distances = self.apply_temperature_correction(timestamp, indices, distances)
        return distances

    @property
    def oversampling(self):
        try: return int(self.yml['configurations']['frequency']['ID_OVERSAMPLING_EXP'])
        except: return 1

    @property
    def binning(self):
        try: return int(self.yml['configurations']['frequency']['binning'])
        except: return 1

    @property
    def base_point_count(self):
        try: return int(self.yml['configurations']['frequency']['ID_BASE_POINT_COUNT'])
        except: bpc = int(self.platform[f'{self.name}_cfg'][0].raw['ID_BASE_POINT_COUNT'])
        return int(bpc/self.binning)

    @property
    def distance_scaling(self):
        """Get the distance in meters between two subsequent points in a trace"""
        return misc.distance_per_sample(self.adc_freq, self.oversampling, self.binning)

    @property
    def trace_smoothing_kernel(self):
        return self.get_trace_smoothing_kernel()

    def get_trace_smoothing_kernel(self):
        """Get the convolution kernel for trace processing for LCA2"""
        kernel = [71,114,71]
        if self.oversampling == 2:
            kernel = [18,34,48,56,48,34,18]
        elif self.oversampling == 4:
            kernel = [6,9,13,17,20,24,26,27,26,24,20,17,13,9,5]
        elif self.oversampling == 8:
            kernel = [4,6,8,11,13,16,18,20,21,22,21,20,18,16,13,11,8,6,4]
        return np.array(kernel)/np.sum(kernel)

    def get_temperature_at(self, timestamp):
        sample = self.datasources['sta'].get_at_timestamp(timestamp, interpolator=floor_interpolator)
        state = sample.raw
        return state['apd_temp'][0]
                 
    def apply_temperature_correction(self, timestamp, indices, distances):
        """Applies temperature-related distance corrections"""
        if (self.temperature_slope is not None and
            self.temperature_reference is not None):
            try:
                temp = self.get_temperature_at(timestamp)
            except Exception as e:
                LoggingManager.instance().warning('Failed to apply temperature correction. '
                                'Error: {}'.format(str(e)))
                return distances

            temp_offset = self.temperature_slope * (temp - self.temperature_reference)
            return distances + temp_offset

        return distances



from pioneer.common import platform
from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.samples import Echo, PointCloud, Image

from typing import Any

import numpy as np


def get_rgb_from_camera_projection(pcloud_ds:PointCloud, camera_ds:str, undistort:bool=False, return_mask:bool=False):
    """Returns the rgb data for each point from its position in camera.
    
        Args:
            pcloud_ds: (Pointcloud) Pointcloud sample (ex: pf["ouster64_bfc_xyzit"][0])
            camera_ds: (str) name of the camera datasource (ex: 'flir_bbfc_flimg')
            undistort: (bool) if True, motion compensation is applied to the points before the projection (default is False)
            return_mask: (bool) if True, also returns the mask that only includes points inside the camera fov.

        Returns:
            rgb: A Nx3 array, where N is the number of points in the point cloud. RGB data is in the range [0,255]
            mask (optional): a Nx1 array of booleans. Values are True where points are inside the camera fov. False elsewhere.
    """

    image_sample:Image = pcloud_ds.datasource.sensor.pf[camera_ds].get_at_timestamp(pcloud_ds.timestamp)

    pcloud = pcloud_ds.get_point_cloud(referential=platform.extract_sensor_id(camera_ds), undistort=undistort)
    projection, mask = image_sample.project_pts(pcloud, mask_fov=True, output_mask=True)
    projection = projection.astype(int)

    rgb = np.zeros((pcloud.shape[0], 3))
    image = image_sample.get_image()
    rgb[mask,:] = image[projection[:,1], projection[:,0]]

    if return_mask:
        return rgb, mask
    return rgb


class RGBCloud(VirtualDatasource):
    """Point cloud with RGB data from the camera projection"""

    def __init__(self, reference_sensor:str, dependencies:list, undistort:bool=False):
        """Constructor
            Args:
                reference_sensor (str): The name of the sensor (e.g. 'pixell_bfc').
                dependencies (list): A list of the datasource names. 
                    The first element should be a point cloud datasource (e.g. 'pixell_bfc_ech')
                    The second element should be a camera image datasource (e.g. 'flir_bfc_img')
                undistort (bool): if True, motion compensentation is applied to the pcloud before the camera projection. 
                    Doesn't affect the point positions, only the rgb data.
        """
        super(RGBCloud, self).__init__(f'xyzit-rgb', dependencies, None)
        self.reference_sensor = reference_sensor
        self.original_pcloud_datasource = dependencies[0]
        self.original_image_datasource = dependencies[1]
        self.camera_name = platform.extract_sensor_id(dependencies[1])
        self.undistort = undistort

        self.dtype = np.dtype([('x','f4'),('y','f4'),('z','f4'),('i','u2'),('t','u8'),('r','u8'),('g','u8'),('b','u8')])

    def __getitem__(self, key:Any):

        if isinstance(key, slice):
            return self[platform.slice_to_range(key, len(self))]
        if isinstance(key, range):
            return [self[index] for index in key]

        pcloud_sample = self.datasources[self.original_pcloud_datasource][key]
        pcloud = pcloud_sample.get_point_cloud(referential=self.camera_name, undistort=self.undistort)
        rgb_data, mask = get_rgb_from_camera_projection(pcloud_sample, self.original_image_datasource, undistort=self.undistort, return_mask=True)

        raw = np.empty(pcloud[mask].shape[0], dtype=self.dtype)
        if isinstance(pcloud_sample, PointCloud):
            raw['x'] = pcloud_sample.get_field('x')[mask]
            raw['y'] = pcloud_sample.get_field('y')[mask]
            raw['z'] = pcloud_sample.get_field('z')[mask]
            raw['i'] = pcloud_sample.get_field('i')[mask]
            raw['t'] = pcloud_sample.get_field('t')[mask]
        elif isinstance(pcloud_sample, Echo):
            # FIXME: Abstraction to remove this part
            pcloud = pcloud_sample.get_point_cloud(ignore_orientation=True)
            raw['x'] = pcloud[:,0][mask]
            raw['y'] = pcloud[:,1][mask]
            raw['z'] = pcloud[:,2][mask]
            raw['i'] = pcloud_sample.amplitudes[mask]
            raw['t'] = pcloud_sample.timestamps[mask]
        raw['r'] = rgb_data[mask,0]
        raw['g'] = rgb_data[mask,1]
        raw['b'] = rgb_data[mask,2]

        return PointCloud(key, self, raw, pcloud_sample.timestamp)


from pioneer.das.api.egomotion import IMUEgomotionProvider
from pioneer.das.api.interpolators import linear_ndarray_interpolator, euler_imu_linear_ndarray_interpolator
from pioneer.das.api.samples import Sample
from pioneer.das.api.sensors.sensor import Sensor



class ImuSbgEkinox(Sensor):

    def __init__(self, name: str, platform: 'Platform'):
        factories = {key: (Sample, linear_ndarray_interpolator) for key in \
        ['gps1vel', 'timeus', 'navposvel', 'status', 'gps1hdt', 'time', 'imudata', 'gps1pos']}
        factories['ekfeuler'] = (Sample, euler_imu_linear_ndarray_interpolator)

        super(ImuSbgEkinox, self).__init__(name, platform, factories)
        
    def create_egomotion_provider(self):
        self.egomotion_provider = IMUEgomotionProvider(self.name, self['navposvel'], self['ekfeuler'])
        return self.egomotion_provider



from pioneer.common                                import linalg
from pioneer.das.api.datasources                   import DatasourceWrapper
from pioneer.das.api.egomotion.egomotion_provider  import EgomotionProvider

import numpy as np
from scipy.interpolate import splprep, splev
import warnings
warnings.simplefilter('once', DeprecationWarning)



class IMUEgomotionProvider(EgomotionProvider):
    """Ego Motion provider for datasets from the RAV-4 platform"""

    def __init__(self, referential_name:str, navposvel:DatasourceWrapper, ekfeuler:DatasourceWrapper):
        super().__init__(referential_name)
        self.navposvel = navposvel
        self.ekfeuler = ekfeuler

    def get_transform(self, timestamp:int) -> np.ndarray:
        nav = self.navposvel.get_at_timestamp(timestamp).raw  # latitude, longitude, altitude
        euler = self.ekfeuler.get_at_timestamp(timestamp).raw # roll, pitch, yaw
        return linalg.tf_from_poseENU(linalg.imu_to_utm(nav, euler), dtype = np.float64) 

    def get_timestamps(self) -> np.ndarray:
        return self.navposvel.timestamps


    ### Legacy section ###

    def trace_sampled_trajectory(self, sampling_distance:float=1.0, N:int=None) -> np.ndarray:
        warnings.warn("IMUEgomotionProvider.trace_sampled_trajectory() is deprecated.", DeprecationWarning)
        xyz = np.empty((0,3))
        for ts in self.navposvel.timestamps:
            xyz = np.vstack([xyz, self.get_Global_from_Ego_at(ts)[:3,3]])

        distances = ((xyz[1:,0] - xyz[:-1,0])**2 + (xyz[1:,1] - xyz[:-1,1])**2)**0.5
        distances = np.insert(distances, 0, 0)
        total_distance = np.sum(distances)

        N = int(total_distance / sampling_distance) + 1 if N is None else N
        
        # To prevent a crash from scipy's splprep, we have to remove any identical consecutive vertices
        keep = np.where(~((xyz[:-1,0] == xyz[1:,0]) & (xyz[:-1,1] == xyz[1:,1])))[0] + 1
        keep = np.insert(keep, 0, 0)

        tck, u = splprep([xyz[keep,0], xyz[keep,1]])
        u = np.linspace(0,1,N)
        return np.array(splev(u, tck)).T



from pioneer.das.api.samples.sample import Sample
from pioneer.das.api import categories
from pioneer.common import platform as platform_utils

from typing import Iterable, Optional

import numpy as np
import warnings
warnings.simplefilter('once', DeprecationWarning)


class Box2d(Sample):
    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super(Box2d, self).__init__(index, datasource, virtual_raw, virtual_ts)

    def get_centers(self) -> np.ndarray:
        return np.vstack((self.raw['data']['x'], self.raw['data']['y'])).T

    def get_dimensions(self) -> np.ndarray:
        return np.vstack((self.raw['data']['h'], self.raw['data']['w'])).T

    def get_confidences(self) -> Iterable[Optional[float]]:
        return self.raw.get('confidence', [None for _ in range(len(self))])

    def get_category_numbers(self) -> Iterable[int]:
        return self.raw['data']['classes']

    def get_categories(self) -> Iterable[str]:
        label_source_name = categories.get_source(platform_utils.parse_datasource_name(self.datasource.label)[2])
        return [categories.CATEGORIES[label_source_name][str(category_number)]['name'] for category_number in self.get_category_numbers()]

    def get_ids(self) -> Iterable[Optional[int]]:
        return self.raw['data']['id']

    def __len__(self) -> int:
        return self.get_centers().shape[0]

    def label_names(self):
        warnings.warn("Box2d.label_names() is deprecated, use Box2d.get_categories() instead.", DeprecationWarning)
        return self.get_categories()



from pioneer.common import platform
from pioneer.common.trace_processing import Realign, TraceProcessingCollection, ZeroBaseline
from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.samples import Trace

from typing import Any

import copy
import numpy as np

class HDRWaveform(VirtualDatasource):
    """High Dynamic Range waveforms are created by combining both sets of waveforms from FastTraces."""

    def __init__(self, reference_sensor:str, dependencies:list, gain:float=1):
        """Constructor
            Args:
                reference_sensor (str): The name of the sensor (e.g. 'pixell_bfc').
                dependencies (list): A list of the datasource names. 
                    The only element should be a FastTrace datasource (e.g. 'pixell_bfc_ftrr')
                gain (float): factor by which the low intensity channel is multiplied before being added to the high intensity channel.
        """
        trr_ds_name = dependencies[0].split('_')[-1].split('-')[-1]
        super(HDRWaveform, self).__init__(f'trr-hdr', dependencies, None)
        self.reference_sensor = reference_sensor
        self.original_trace_datasource = dependencies[0]
        self.gain = gain

        self.trace_processing = TraceProcessingCollection([ZeroBaseline(), Realign()])
        self.target_time_base_delay = None

    def __getitem__(self, key:Any):

        if isinstance(key, slice):
            return self[platform.slice_to_range(key, len(self))]
        if isinstance(key, range):
            return [self[index] for index in key]

        trace_sample = self.datasources[self.original_trace_datasource][key]
        timestamp = trace_sample.timestamp

        if self.target_time_base_delay is None:
            self.target_time_base_delay = max([
                np.max(trace_sample.raw['high']['time_base_delays']),
                np.max(trace_sample.raw['low']['time_base_delays']),
            ])
            self.trace_processing = TraceProcessingCollection([ZeroBaseline(), Realign(self.target_time_base_delay)])

        processed_traces = trace_sample.processed(self.trace_processing)

        raw = copy.deepcopy(processed_traces['high'])

        low_traces = processed_traces['low']['data']
        raw['data'][:,:low_traces.shape[-1]] += self.gain*low_traces

        return Trace(key, self, raw, timestamp)